import { FunctionHandler, IntegrationHandler } from "@netlify/integrations";
import {
  captureException,
  captureMessage,
  AWSLambda,
  NodeOptions,
  configureScope,
} from "@sentry/serverless";
import { basename } from "path";
import { HandlerEvent } from "@netlify/functions";
import { ampli } from "ampli/index.js";

export type NetlifySentryConfig = {
  cronMonitoring?: {
    enable: boolean;
    monitorId: string;
  };
};

export type SentryConfig = {
  sentry?: NodeOptions & NetlifySentryConfig;
};

export interface SentryContext {
  sentry: {
    captureException: typeof captureException;
    captureMessage: typeof captureMessage;
  };
}

export const getNetlifyConfig = (
  sentryConfig: SentryConfig["sentry"] = {}
): NetlifySentryConfig => {
  const netlifySentryConfigKeys = ["cronMonitoring"];
  const netlifyConfig: {
    [key: string]: unknown;
  } = {};

  // Removes configuration not part of the Sentry SDK init before returning
  netlifySentryConfigKeys.forEach((key) => {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    netlifyConfig[key] = (sentryConfig as any)[key];
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    delete (sentryConfig as any)[key];
  });

  return netlifyConfig;
};

const getCronMonitoringCheckInRequestHeaders = () => {
  const requestHeaders = new Headers();
  requestHeaders.set("Authorization", `DSN ${process.env.SENTRY_DSN}`);
  requestHeaders.set("Content-Type", "application/json");

  return requestHeaders;
};

export const startSentryCheckIn = async (monitorId: string) => {
  // Todo: replace with datadog monitoring
  console.log("Starting Sentry check in");
  const requestHeaders = getCronMonitoringCheckInRequestHeaders();

  await fetch(`https://sentry.io/api/0/monitors/${monitorId}/checkins/`, {
    method: "POST",
    headers: requestHeaders,
    body: JSON.stringify({ status: "in_progress" }),
  });
};

export const completeSentryCheckIn = async (monitorId: string) => {
  const requestHeaders = getCronMonitoringCheckInRequestHeaders();

  await fetch(
    `https://sentry.io/api/0/monitors/${monitorId}/checkins/latest/`,
    {
      method: "PUT",
      headers: requestHeaders,
      body: JSON.stringify({ status: "ok" }),
    }
  );

  // Todo: replace with datadog monitoring
  console.log("Sentry check in completed");
};

const initializeAmplitude = async () => {
  if (!ampli.isLoaded) {
    const environment =
      process.env.NODE_ENV === "production" ? "production" : "development";

    await ampli.load({ environment }).promise;
  }
};

export const withSentry: IntegrationHandler<SentryContext, SentryConfig> = (
  handler,
  config?
) => {
  const sentry = {
    captureException,
    captureMessage,
  };
  const { cronMonitoring } = getNetlifyConfig(config?.sentry);
  const sentryConfigUserOverrides: NodeOptions = config?.sentry || {};

  AWSLambda.init({
    dsn: process.env.SENTRY_DSN,
    // This value means that all events are sent to Sentry. While
    // this is not ideal for functions with higher numbers of invocations
    // it will ensure that those just getting started will see their events
    // be sent immediately to Sentry
    tracesSampleRate: 1.0,
    ...sentryConfigUserOverrides,
  });

  return AWSLambda.wrapHandler(async (event: HandlerEvent, context) => {
    await initializeAmplitude();

    const functionName = `netlify/functions/${basename(__filename)}`;
    ampli.integrationSentryEnabled(process.env.SITE_ID, { functionName });

    if (cronMonitoring?.enable) {
      ampli.integrationSentryCronMonitoringEnabled(process.env.SITE_ID, {
        functionName,
      });
      await startSentryCheckIn(cronMonitoring.monitorId);
    }

    configureScope((scope) => scope.setTransactionName(functionName));

    if (cronMonitoring?.enable) {
      return handler(event, { ...context, sentry }).finally(async () => {
        await completeSentryCheckIn(cronMonitoring.monitorId);
      });
    }

    // Clear the Amplitude event buffer before invoking the user's handler method
    // to ensure they're all sent
    await ampli.flush().promise;

    return handler(event, { ...context, sentry });
  }) as FunctionHandler;
};
