import {
  vi,
  test,
  describe,
  beforeEach,
  afterEach,
  expect,
  Mock,
} from "vitest";
import {
  SentryContext,
  SentryConfig,
  withSentry,
  getNetlifyConfig,
  startSentryCheckIn,
  completeSentryCheckIn,
  // Ignore this rule for the time being.
  // See https://github.com/microsoft/TypeScript/issues/49083
  // This was still an issue as of TS v4.9, hopefully gets sorted as 
  // the current 'fix' is to add a '.js' extension which doesn't make sense
  // since that file doesn't actually exist.
  // @ts-ignore TS2835
} from "../src/index";
import Chance from "chance";
import { HandlerContext, HandlerEvent } from "@netlify/functions";
import { FunctionHandler } from "@netlify/integrations";

const chance = new Chance();

/**
 * This is the same as what's in 'getCronMonitoringCheckInRequestHeaders' but duplicated here
 * in order to:
 *
 * 1) Make it easy for us to construct request headers for mock requests
 * 2) Ensure that if someone accidentally modifies the request headers in the other function
 * that it's caught in these tests
 */
const constructSentryRequestHeaders = () => {
  const requestHeaders = new Headers();
  requestHeaders.set("Authorization", `DSN ${process.env.SENTRY_DSN}`);
  requestHeaders.set("Content-Type", "application/json");

  return requestHeaders;
};

describe("getNetlifyConfig", () => {
  let handlerConfiguration: SentryConfig;

  beforeEach(() => {
    handlerConfiguration = {
      sentry: {
        tracesSampleRate: chance.floating({ fixed: 1, min: 0, max: 1 }),
        cronMonitoring: {
          enable: true,
          monitorId: chance.guid(),
        },
      },
    };
  });

  test("does not modify Sentry configuration", () => {
    const originalValue = handlerConfiguration.sentry?.tracesSampleRate;

    getNetlifyConfig(handlerConfiguration.sentry);
    expect(handlerConfiguration.sentry?.tracesSampleRate).toEqual(
      originalValue
    );
  });

  test("removes Netlify Sentry configuration from the user-provided configuration object", () => {
    getNetlifyConfig(handlerConfiguration.sentry);
    expect(handlerConfiguration.sentry?.cronMonitoring).not.toBeDefined();
  });

  test("returns the Netlify Sentry configuration", () => {
    const originalCronMonitoringConfig = {
      cronMonitoring: handlerConfiguration.sentry?.cronMonitoring,
    };

    const netlifySentryConfig = getNetlifyConfig(handlerConfiguration.sentry);

    expect(netlifySentryConfig).toStrictEqual(originalCronMonitoringConfig);
  });
});

describe("startSentryCheckIn", () => {
  beforeEach(() => {
    process.env.SENTRY_DSN = `https://${chance.string({
      alpha: true,
      numeric: true,
    })}@${chance.string({
      alpha: true,
      numeric: true,
    })}.ingest.sentry.io/${chance.integer({ min: 0 })}`;
  });

  afterEach(() => {
    vi.restoreAllMocks();
    delete process.env.SENTRY_DSN;
  });

  test("makes the request with the correct headers", async () => {
    const spy = vi.fn();

    global.fetch = spy;

    const monitorId = chance.guid();
    await startSentryCheckIn(monitorId);

    const expectedUrl = `https://sentry.io/api/0/monitors/${monitorId}/checkins/`;
    const expectedPayload = {
      method: "POST",
      headers: constructSentryRequestHeaders(),
      body: JSON.stringify({ status: "in_progress" }),
    };
    expect(spy).toHaveBeenCalledWith(expectedUrl, expectedPayload);
  });
});

describe("completeSentryCheckIn", () => {
  beforeEach(() => {
    process.env.SENTRY_DSN = `https://${chance.string({
      alpha: true,
      numeric: true,
    })}@${chance.string({
      alpha: true,
      numeric: true,
    })}.ingest.sentry.io/${chance.integer({ min: 0 })}`;
  });

  afterEach(() => {
    vi.restoreAllMocks();
    delete process.env.SENTRY_DSN;
  });

  test("makes the request with the correct headers", async () => {
    const spy = vi.fn();

    global.fetch = spy;

    const monitorId = chance.guid();
    await completeSentryCheckIn(monitorId);

    const expectedUrl = `https://sentry.io/api/0/monitors/${monitorId}/checkins/latest/`;
    const expectedPayload = {
      method: "PUT",
      headers: constructSentryRequestHeaders(),
      body: JSON.stringify({ status: "ok" }),
    };
    expect(spy).toHaveBeenCalledWith(expectedUrl, expectedPayload);
  });
});

describe("withSentry", () => {
  let fetchSpy: Mock;
  let mockHandler: FunctionHandler;

  beforeEach(() => {
    vi.mock("@sentry/serverless", async (importOriginal) => {
      // eslint-disable-next-line @typescript-eslint/no-explicit-any
      const mod: any = await importOriginal();
      const wrapHandler = mod.AWSLambda.wrapHandler;

      return {
        ...mod,
        AWSLambda: {
          init: vi.fn(),
          wrapHandler,
        },
        configureScope: vi.fn(),
      };
    });
    fetchSpy = vi.fn();

    global.fetch = fetchSpy;
    process.env.SENTRY_DSN = `https://${chance.string({
      alpha: true,
      numeric: true,
    })}@${chance.string({
      alpha: true,
      numeric: true,
    })}.ingest.sentry.io/${chance.integer({ min: 0 })}`;

    mockHandler = async (
      /* eslint-disable @typescript-eslint/no-unused-vars */
      event: HandlerEvent,
      context: HandlerContext & SentryContext
      /* eslint-enable @typescript-eslint/no-unused-vars */
    ) => {
      return {
        statusCode: 200,
      };
    };
  });

  afterEach(() => {
    vi.restoreAllMocks();
    vi.resetModules();
    delete process.env.SENTRY_DSN;
  });

  test("it performs Sentry check ins when cron monitoring is enabled", async () => {
    const monitorId = chance.guid();
    const mockConfig = {
      sentry: {
        cronMonitoring: {
          enable: true,
          monitorId,
        },
      },
    };

    const res = withSentry(mockHandler, mockConfig);
    await res({} as HandlerEvent, {} as HandlerContext);

    expect(fetchSpy).toHaveBeenCalledTimes(2);
    expect(fetchSpy.mock.calls[0][0]).toEqual(
      `https://sentry.io/api/0/monitors/${monitorId}/checkins/`
    );
    expect(fetchSpy.mock.calls[1][0]).toEqual(
      `https://sentry.io/api/0/monitors/${monitorId}/checkins/latest/`
    );
  });

  test("it does not perform Sentry check ins when cron monitoring is disabled", async () => {
    const monitorId = chance.guid();
    const mockConfig = {
      sentry: {
        cronMonitoring: {
          enable: false,
          monitorId,
        },
      },
    };

    const res = withSentry(mockHandler, mockConfig);
    await res({} as HandlerEvent, {} as HandlerContext);

    expect(fetchSpy).not.toBeCalled();
  });

  test("it calls configureScope in order to set the transaction name to the filename of the function", async () => {
    const configureScopeMock = (await import("@sentry/serverless"))
      .configureScope;

    const monitorId = chance.guid();
    const mockConfig = {
      sentry: {
        cronMonitoring: {
          enable: false,
          monitorId,
        },
      },
    };

    const res = withSentry(mockHandler, mockConfig);
    await res({} as HandlerEvent, {} as HandlerContext);

    expect(configureScopeMock).toBeCalled();
  });
});
